"""Simple MLP PyTorch models."""

import torch


class MLP(torch.nn.Module):
    """Basic multi-layer perceptron."""

    def __init__(
        self,
        in_features: int,
        out_features: int,
        hidden_features: int = 128,
        n_hidden_layers: int = 3,
        activation: type = torch.nn.LeakyReLU,
    ):
        super().__init__()

        modules = [
            torch.nn.Linear(in_features=in_features, out_features=hidden_features),
        ]
        if activation is not None:
            modules.append(activation())
        for _ in range(n_hidden_layers):
            modules.append(
                torch.nn.Linear(
                    in_features=hidden_features, out_features=hidden_features
                )
            )
            if activation is not None:
                modules.append(activation())
        modules.append(
            torch.nn.Linear(in_features=hidden_features, out_features=out_features)
        )
        self.module_list = torch.nn.ModuleList(modules)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        for layer in self.module_list:
            x = layer(x)
        return x


if __name__ == "__main__":
    # Basic MLP usage example.
    in_features = 1
    out_features = 1
    network = MLP(in_features=in_features, out_features=out_features)
    batch_size = 8
    out = network(torch.ones((batch_size, in_features)))



"""Dataset for simulated polynomial data."""

from collections.abc import Callable
from typing import Optional

import torch


class Polynomial(torch.utils.data.Dataset):
    """Dataset for stochastic simulators with fixed input arguments."""

    def __init__(
        self,
        data_generator: Callable,
        dataset_length: int = 1,
        cache: bool = False,
        transform: Optional[torch.nn.Module] = None,
    ) -> None:
        """Initializer for stochastic simulator dataset.

        Args:
            data_generator:  Callable that can be called to generate data as data = data_generator()
            dataset_length: Desired dataset length (set manually since `simulator` is stochastic).
            transform: Transform applied to simulated data before returning.
        """
        super().__init__()
        self.data_generator = data_generator
        self._len = dataset_length
        self.transform = transform
        self.cache = cache
        self.data = [None for _ in range(dataset_length)]

    def __len__(self) -> int:
        return self._len

    def __getitem__(self, idx: int) -> tuple:
        if self.cache and self.data[idx] is not None:
            return self.data[idx]
        else:
            data_dict = self.data_generator()
            if self.transform is not None:
                data_dict["output"] = self.transform(data_dict["output"])
            if self.cache:
                self.data[idx] = (data_dict["input"]["x"], data_dict["output"])

        return data_dict["input"]["x"], data_dict["output"]



"""Definitions of prior distributions over neural network parameters."""

from collections.abc import Iterable
import copy
from typing import Any, Optional

import torch
import torch.distributions as dist


class Distribution(torch.nn.Module):
    """Distribution wrapper to facilitate device transfers."""

    def __init__(
        self,
        distribution: Optional[dist.Distribution] = None,
        *args: Any,
        **kwargs: Any
    ):
        """Initialize wrapper."""
        super().__init__()

        # If `distribution` is passed, we'll build the wrapper automatically.
        if distribution is not None:
            self._distribution = copy.deepcopy(distribution)
            for key, val in distribution.__dict__.items():
                if isinstance(val, torch.Tensor):
                    self.register_buffer(key, val)

    @property
    def distribution(self) -> dist.Distribution:
        """Prepare an instance of the distribution."""
        for key, val in self._distribution.__dict__.items():
            if isinstance(val, torch.Tensor):
                buffer_val = getattr(self, key)
                setattr(self._distribution, key, buffer_val)
        return self._distribution

    def log_prob(self, x: torch.Tensor) -> torch.Tensor:
        """Compute the log PDF of the prior at points `x`."""
        return self.distribution.log_prob(x)

    def sample(self, sample_shape: Iterable = torch.Size()) -> torch.Tensor:
        """Generate samples from the prior of size `sample_shape`."""
        return self.distribution.sample(sample_shape=sample_shape)


class SpikeSlab(Distribution):
    """Spike-slab Gaussian Mixture Model prior."""

    def __init__(
        self,
        loc: torch.Tensor = torch.tensor([0.0, 0.0]),
        scale: torch.Tensor = torch.tensor([0.1, 1.0]),
        probs: torch.Tensor = torch.tensor([0.5, 0.5]),
    ):
        super().__init__()

        self.register_buffer("loc", loc)
        self.register_buffer("scale", scale)
        self.register_buffer("probs", probs)

    @property
    def distribution(self) -> dist.Distribution:
        """Prepare an instance of the distribution."""
        mixture_distribution = dist.Categorical(probs=self.probs)
        return dist.MixtureSameFamily(
            mixture_distribution=mixture_distribution,
            component_distribution=dist.Normal(loc=self.loc, scale=self.scale),
        )


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    # Example using an SpikeSlab prior.
    prior = SpikeSlab()
    sample = prior.sample(sample_shape=(100, 1))
    x = torch.linspace(-5.0, 5.0, 1000)
    pdf = torch.exp(prior.log_prob(x))
    fig, ax = plt.subplots()
    ax.hist(sample, density=True)
    ax.plot(x, pdf)
    plt.show()



"""Collections of Bayesian neural network layers."""

from abc import ABC, abstractmethod
import copy
import functools
import re
import sys
from typing import Any, Optional, Union

import numpy as np
import torch
import torch.distributions as dist

from . import inference
from .inference import MomentPropagator
from .losses import kl_divergence_sampled
from .priors import Distribution
from .types import MuVar


# Define layers that can be applied to input mean and variance without additional
# processing (e.g., a flatten layer, which only changes the shape of the input).
BROADCAST = [
    "ChannelShuffle",
    "Identity",
    "Flatten",
    "Unflatten",
    *[f"ReflectionPad{n + 1}d" for n in range(3)],
    *[f"ReplicationPad{n + 1}d" for n in range(3)],
    *[f"ZeroPad{n + 1}d" for n in range(3)],
    *[f"ConstantPad{n + 1}d" for n in range(3)],
    *[f"CircularPad{n + 1}d" for n in range(3)],
]


CURRENT_MODULE = sys.modules[__name__]


def select_default_propagator(
    module: torch.nn.Module, is_bayesian: bool = True
) -> MomentPropagator:
    """Select a compatible moment propagator for `module`.

    Args:
        module: Module that we'll choose a propagator for.
        is_bayesian: Flag indicating parameters of `module` will be treated
            as distributions.
    """
    if hasattr(inference, module.__class__.__name__):
        # A custom propagator exists for this module so we'll use that.
        propagator = getattr(inference, module.__class__.__name__)
        moment_propagator = propagator()
    elif not is_bayesian or (
        len([p for p in module.parameters() if p.requires_grad]) == 0
    ):
        # If this module doesn't have learnable parameters we'll
        # default to the unscented transform.
        moment_propagator = inference.UnscentedTransform()
    else:
        # With learnable Bayesian parameters, we'll default to
        # Monte Carlo sampling.
        moment_propagator = inference.MonteCarlo()

    return moment_propagator


def isolate_leaf_module_names(module_names: list[str]) -> list[str]:
    """Prepare a list of leaf modules of `model`.

    This function filters `module_names` to eliminate the names of parent modules.
    For example, if we have a model: torch.nn.Module with named modules
    m = ["", "module1", "module2", "module1.submodule", "module2.submodule"],
    isolate_leaf_module_names(m) == ["module1.submodule", "module2.submodule"]
    """
    leaf_names = []
    module_names.remove("")  # remove root module empty string
    for module in module_names[::-1]:
        # If other modules are a prefix of this modules name, we'll assume they
        # are this modules parent (hence not a leaf module).
        children = []
        for leaf in leaf_names:
            matches = re.match(f"{module}\..*", leaf)
            if matches is not None:
                children.append(matches)
        if len(children) == 0:
            leaf_names.append(module)

    return leaf_names


def convert_to_bnn_(
    model: torch.nn.Module,
    layer_wrappers: dict = {},
    wrapper_kwargs: dict = {},
    wrapper_kwargs_global: dict = {},
    broadcast_module_tags: Union[list, tuple] = (),
) -> None:
    """Convert layers of `model` to Bayesian counterparts.

    Args:
        model: Model to be converted to Bayesian counterpart.
        layer_wrappers: Dictionary of manually-specified wrappers for specific
            module layers. The keys are names of leaf modules (e.g.,
            "module1.layer1") and the values are the corresponding wrapper
            class present in this module (e.g., "BroadcastModule").
        wrapper_kwargs: Additional keyword arguments passed to
            initialization of named Bayesian layers.  For example, if `model`
            has a module named "module1", we'll convert "module1" as
            Converter(module1, **wrapper_kwargs["module1"]) where
            Converter is a module converter.
        wrapper_kwargs_global: Keyword arguments that we'll merge
            with values of wrapper_kwargs as, e.g.,
            Converter(module1, **(wrapper_kwargs_global | wrapper_kwargs["module1"]))
        broadcast_module_tags: List of strings that, if present in the class
            name of a module, will indicate the module should be treated as a
            broadcast module, i.e., apply forward method to mean and variance
            directly without additional logic.
    """
    # Search for modules of `model` to convert, removing stem modules from the
    # list (we just want the leaf modules that contain parameters).
    module_names = [n for n, _ in model.named_modules()]
    leaf_names = isolate_leaf_module_names(module_names)

    # Replace leaf modules with Bayesian counterparts or compatible passthroughs.
    for leaf in leaf_names:
        module = model.get_submodule(leaf)
        module_name = module.__class__.__name__

        # Prepare module arguments.
        module_kwargs = wrapper_kwargs_global | wrapper_kwargs.pop(leaf, {})

        # Search for an appropriate module converter, in the following order of
        # priority:
        #   (1) User-specified wrapper designated in `layer_wrappers`.
        #   (2) Broadcast layer if tagged by broadcast_module_tags or listed
        #       in PASSTHROUGH list.
        #   (3) Named converters if a wrapper exists with the same name as the
        #       module class.
        #   (4) BayesianModule
        if wrapper := layer_wrappers.pop(leaf, None):
            # Wrap module in user-specified wrapper.
            bayesian_layer = getattr(CURRENT_MODULE, wrapper)(
                module=module, **module_kwargs
            )
        elif (module_name in BROADCAST) or any(
            [tag in module_name for tag in broadcast_module_tags]
        ):
            # This module can be broadcast along (mu, var) without additional
            # processing (e.g., a flatten layer, which only changes shapes).
            bayesian_layer = BroadcastModule(module=module, **module_kwargs)
        elif hasattr(CURRENT_MODULE, module_name):
            # If a custom converter exists for this named layer, we'll use that by default.
            bayesian_layer = getattr(CURRENT_MODULE, module_name)(
                module=module, **module_kwargs
            )
        else:
            bayesian_layer = BayesianModule(module=module, **module_kwargs)

        # Reset submodule to the converted module.
        model.set_submodule(leaf, bayesian_layer)


def convert_to_nn(
    bnn: torch.nn.Module,
) -> None:
    """Inverse of convert_to_bnn_ to convert a BNN back to a standard NN

    Args:
        model: Bayesian NN to be converted back to a standard NN.
    """
    # Search for modules of `model` to convert, removing stem modules from the
    # list (we just want the leaf modules that contain parameters).
    model = copy.deepcopy(bnn)
    module_names = [n for n, _ in model.named_modules()]
    leaf_names = isolate_leaf_module_names(module_names)

    # Remove BNN-specific modules from list.
    bnn_module_names = ["_module_params", "_moment_propagator"]
    leaf_names = [
        leaf for leaf in leaf_names if not any([bn in leaf for bn in bnn_module_names])
    ]

    # Replace Bayesian leaf modules with standard counterparts.
    for leaf in leaf_names:
        # If `leaf` is a named `_module` parameter, we'll reset the module to the `_module` leaf.
        # If the module is a BroadcastModule, we just need to remove the wrapper.
        module = model.get_submodule(leaf)
        leaf_split = leaf.split(".")
        if leaf_split[-1] == "_module":
            model.set_submodule(".".join(leaf_split[:-1]), module)
        elif isinstance(module, BroadcastModule):
            model.set_submodule(leaf, module.module)

    return model


class BayesianModuleBase(ABC, torch.nn.Module):
    """Abstract base class for Bayesian modules."""

    # Define a scale transform to transform learnable `rho` parameters to scale
    # parameters (e.g., sigma = torch.nn.functional.softplus(rho)) where `rho` is
    # learned and `sigma` is the standard deviation of the Gaussian defining
    # a module's parameters.
    scale_tform = torch.nn.functional.softplus

    def scale_tform_inv(x: torch.Tensor) -> torch.Tensor:
        # Inverse of self.scale_tform.
        return torch.log(torch.exp(x) - 1.0)

    @property
    @abstractmethod
    def learn_var(self, *args, **kwargs) -> bool:
        """Flag indicating variance of module parameters should be learnable."""
        pass

    @property
    @abstractmethod
    def module_params(self, *args, **kwargs) -> torch.ParameterDict:
        """Dictionary organizing parameters of this module."""
        pass

    @property
    @abstractmethod
    def module(self, *args, **kwargs) -> torch.nn.Module:
        """Return a sampled instance of this module."""
        pass

    @property
    @abstractmethod
    def module_map(self, *args, **kwargs) -> torch.nn.Module:
        """Return a non-Bayesian instance of self with parameters set to learned means."""
        pass

    @property
    @abstractmethod
    def moment_propagator(self, *args, **kwargs) -> MomentPropagator:
        """Function used to propagate mean and variance through this module."""
        pass

    @abstractmethod
    def compute_kl_divergence(
        self,
        *args,
        **kwargs,
    ) -> torch.Tensor:
        """Method that computes the KL divergence of this module w.r.t. some prior."""
        pass

    @abstractmethod
    def forward(
        self,
        input: Union[MuVar, torch.Tensor],
        *args,
        **kwargs,
    ) -> Union[MuVar, torch.Tensor]:
        """Method that computes a forward pass through this module."""
        pass


class BayesianModule(BayesianModuleBase):
    """Base class for BayesianModule modules to make PyTorch modules BNN compatible."""

    def __init__(
        self,
        module: torch.nn.Module,
        samplers: Optional[dict] = None,
        samplers_init: Optional[dict] = None,
        resample_mean: bool = True,
        priors: Optional[dict] = None,
        moment_propagator: Optional[MomentPropagator] = None,
        learn_var: bool = True,
        *args,
        **kwargs,
    ) -> None:
        """BayesianModule initializer.

        Args:
            module: Module with parameters that we wish to treat as distributions
                (or in some cases, a module between Bayesian layers that we'll
                wrap for compatibility purposes).
            samplers: Dictionary of distributions that are sampled to return
                parameters of this layer (i.e., the variational distributions).
                These should be uninitialized or partially initialized
                distributions that accept inputs `loc` and `scale` and return
                an object with a .rsample() method which returns tensors
                corresponding to sampled parameters.  For example,
                samplers["weight"](loc=0.0, scale=1.0).rsample()
                must return a tensor of size (out_features, in_features) if
                module = torch.nn.Linear(in_features, out_features)
            samplers_init: Dictionary of samplers that can be sampled to reset
                parameters of this layer.  Each value is a tuple corresponding
                to the mean and (unscaled, see `scale_tform` usage) standard deviation
                parameters (e.g., samplers_init["weight_mean"].rsample() should
                return a shape (out_features, in_features) tensor defining
                initial weight parameters).  Keys should match those in `samplers`.
            resample_mean: Flag indicating parameter means should be resampled using
                appropriate samplers from `samplers_init` before training. This can
                be set to False to initialize mean values to parameter values of the
                input module (e.g., if pretrained).
            priors: Dictionary of prior distributions over layer parameters.
                These distributions should be initialized (like `samplers_init`)
                and have a .rsample() method returning a tensor of the same
                size as the corresponding layer parameter and a .log_prob(x) method
                returning the log of the PDF at input x in the distributions domain.
                These distributions are used in self.compute_kl_divergence().
                Keys should match those in `samplers`.
            moment_propagator: Propagation module to propagate mean and variance
                through this layer.
            learn_var: Flag indicating learnable parameters should be treated as
                distributions (with learned variance).  This flag allows us to
                wrap a layer with this module and its functionality without
                changing the behavior of its parameters.
        """
        super().__init__(*args, **kwargs)
        self.__name__ = module.__class__.__name__

        # Define mean and unscaled standard deviation parameters.
        module_params = [p for p in module.named_parameters()]
        _module_params = torch.nn.ParameterDict()
        self._learn_var = learn_var
        if (len(module_params) > 0) and learn_var:
            for key, _ in module_params:
                _module_params[key + "_mean"] = getattr(module, key)
                _module_params[key + "_rho"] = torch.nn.Parameter(
                    torch.empty_like(_module_params[key + "_mean"])
                )
        else:
            for key, _ in module_params:
                _module_params[key + "_mean"] = getattr(module, key)
                _module_params[key + "_rho"] = None
        self._module = module
        self._module_params = _module_params

        # Validate and set moment propagator.
        # Set moment propagator.
        if moment_propagator is None:
            moment_propagator = select_default_propagator(
                module=module, is_bayesian=learn_var
            )
        self._moment_propagator = moment_propagator

        # Create samplers for each named parameter.
        if samplers_init is None:
            # Prepare (relatively arbitrary) default samplers.
            samplers_init = {}
            for key, val in _module_params.items():
                if "_mean" in key:
                    samplers_init[key] = dist.Uniform(
                        low=-1.0 / np.sqrt(val.shape[-1]),
                        high=1.0 / np.sqrt(val.shape[-1]),
                    )
                else:
                    samplers_init[key] = dist.Uniform(
                        low=-8.0,
                        high=-2.0,
                    )
        self.samplers_init = samplers_init
        self.resample_mean = resample_mean
        if learn_var:
            self.reset_parameters()

        # Define variational distribution sampler types.
        if samplers is None:
            samplers = {}
            for key, _ in module.named_parameters():
                if self._module_params[key + "_rho"] is None:
                    samplers[key] = None
                else:
                    samplers[key] = dist.Normal
        self._samplers = samplers

        # Set priors.
        if priors is None:
            # Default to same distributions as `samplers_init` mean samplers.
            priors = {
                key: Distribution(samplers_init[key + "_mean"])
                for key in samplers.keys()
            }
        self.priors = priors

    @property
    def learn_var(self) -> bool:
        """Return property `learn_var`.

        This property is written as an @property method for compatibility with the
        abstract parent class.
        """
        return self._learn_var

    @property
    def moment_propagator(self) -> bool:
        """Return property `moment_propagator`.

        This property is written as an @property method for compatibility with the
        abstract parent class.
        """
        return self._moment_propagator

    def get_named_sampler(self, name: str) -> torch.distributions.Distribution:
        """Initialize and return the requested module parameter sampler."""
        if self._samplers[name] is None:
            return None
        else:
            return self._samplers[name](
                loc=self._module_params[name + "_mean"],
                scale=self.scale_tform(self._module_params[name + "_rho"]),
            )

    @property
    def samplers(self) -> dict:
        """Return initialized variational distribution samplers."""
        # Initialize the samplers on the fly to ensure we're on the same device
        # as self._module_params.
        return {key: self.get_named_sampler(name=key) for key in self._samplers.keys()}

    @property
    def module_map(self) -> torch.nn.Module:
        """Return module instance with parameters set to learned means."""
        return self._module

    @property
    def module_params(self) -> dict:
        """Return a sample of this module's parameters."""
        params = {}
        for key, val in self.samplers.items():
            if val is None:
                # This parameter is not treated as a distribution so we can
                # directly return the mean value.
                params[key] = self._module_params[key + "_mean"]
            else:
                params[key] = val.rsample()

        return params

    @property
    def module(self) -> torch.nn.Module:
        """Prepare a callable that acts like input `module` with random parameters."""
        return functools.partial(
            torch.func.functional_call, self._module, self.module_params
        )

    def __getattr__(self, name: str) -> Any:
        """Custom getattr to return samples of named parameters."""
        # Check for a sampler associated with `name` and return a sample if it exists.
        inst_modules = self.__dict__.get("_modules", {})
        if f"{name}_mean" in inst_modules.get("_module_params", {}).keys():
            return self.get_named_sampler(name=name).sample()

        # Check remaining modules and parameters for requested attribute.
        inst_params = self.__dict__.get("_parameters", {})
        if name in inst_modules.keys():
            return inst_modules[name]
        elif name in inst_params.keys():
            return inst_params[name]

        # Raise attribute error if we reach this point.
        raise AttributeError(
            f"'{type(self).__name__}' object has no attribute '{name}'"
        )

    def reset_parameters(self) -> None:
        """Resample layer parameters from initial distributions."""
        for key, param in self._module_params.items():
            if param is not None:
                # If this is a parameter mean, verify self.resample_mean flag
                # before resampling.
                if ("_mean" not in key) or self.resample_mean:
                    param.data = self.samplers_init[key].sample(
                        sample_shape=param.shape
                    )

    def compute_kl_divergence(
        self, priors: Optional[Union[dict, Distribution]] = None, n_samples: int = 1
    ) -> torch.Tensor:
        """Compute the KL divergence between self.prior and module parameters.

        Args:
            priors: Prior distribution over parameters.  This can be a single
                distribution for all parameters or a dictionary whose keys
                correspond to named parameters of this layer.  By default, None
                will use self.priors.  This argument is used to allow external
                passing of priors not defined at initialization of this layer.
            n_samples: Number of Monte Carlo samples used to estimate KL
                divergence for distributions not compatible with
                torch.nn.distributions.kl_divergence()
        """
        # Reorganize priors if needed.
        if priors is None:
            priors = self.priors
        elif not isinstance(priors, dict):
            priors = {key: priors for key in self._samplers.keys()}

        # Loop over parameters and compute KL contribution.
        kl_divergence = []
        for param_name, param_prior in priors.items():
            # Compute KL divergence for this parameter.
            try:
                # Attempt to use PyTorch kl_divergence.  If our distributions
                # aren't compatible, dist.kl_divergence kicks us to the except
                # clause below.
                kl_divergence.append(
                    dist.kl_divergence(
                        self.get_named_sampler(name=param_name),
                        param_prior.distribution,
                    ).sum()
                )
            except NotImplementedError:
                # Compute Monte Carlo KL divergence.
                kl_divergence.append(
                    kl_divergence_sampled(
                        dist0=self.get_named_sampler(name=param_name),
                        dist1=param_prior.distribution,
                        n_samples=n_samples,
                    ).sum()
                )

        return torch.stack(kl_divergence).sum()

    def forward(
        self,
        input: Union[MuVar, torch.Tensor],
        *args,
        **kwargs,
    ) -> Union[MuVar, torch.Tensor]:
        """Forward pass through layer."""
        if isinstance(input, MuVar):
            # If the input has no variance, set to zero before propagating.
            if input[1] is None:
                input = MuVar(input[0], torch.zeros_like(input[0]))

            # Propagate mean and variance through layer.
            out = self.moment_propagator(
                module=self,
                input=input,
                *args,
                **kwargs,
            )
        else:
            # Compute forward pass with a random sample of parameters.
            out = self.module(input, *args, **kwargs)

        return out


class BroadcastModule(torch.nn.Module):
    """BroadcastModule for compatibility with other Bayesian layers."""

    def __init__(
        self,
        module: torch.nn.Module,
        *args,
        **kwargs,
    ) -> None:
        """BroadcastModule initializer.

        Args:
            module: Module that will be applied in forward pass.  When the
                output of the previous layer is a MuVar type, module is
                simply broadcast across both elements of MuVar.
        """
        super().__init__()
        self.__name__ = module.__class__.__name__
        self.module = module

    def forward(
        self,
        input: Union[MuVar, torch.Tensor],
        *args,
        **kwargs,
    ) -> Union[MuVar, torch.Tensor]:
        """Forward pass through layer."""
        if isinstance(input, MuVar):
            # Propagate mean and variance through layer.
            if input[1] is None:
                # No input variance so we only need to operate on mean.
                out = MuVar(self.module(input[0]), None)
            else:
                out = MuVar(self.module(input[0]), self.module(input[1]))
        else:
            # Compute forward pass with a random sample of parameters.
            out = self.module(input)

        return out


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    from inference import Linear

    # Basic usage example of BayesianLinear.
    in_features = 3
    out_features = 1
    linear = torch.nn.Linear(in_features=in_features, out_features=out_features)
    bayesian_linear = BayesianModule(
        module=linear,
    )

    # Usage with MonteCarlo moment propagation through layer.
    n_samples = 100
    bayesian_linear_mc = BayesianModule(
        module=linear,
        moment_propagator=inference.MonteCarlo(n_samples=n_samples),
    )  # computes moments from Monte Carlo without returning actual samples
    bayesian_linear_mc.load_state_dict(bayesian_linear.state_dict())

    # Usage with analytical moment propagation through layer.
    bayesian_linear_det = BayesianModule(
        module=linear,
        moment_propagator=Linear(),
    )  # analytically computes moments
    bayesian_linear_det.load_state_dict(bayesian_linear.state_dict())

    # Generate example outputs.  Wrapping input in type MuVar will signal the
    # layers to use their moment_propagator modules instead of sampling.
    batch_size = 1
    input = torch.randn(batch_size, in_features)
    output_samples = torch.stack([bayesian_linear(input)[0] for _ in range(n_samples)])
    output_mc = bayesian_linear_mc(MuVar(input))
    output_det = bayesian_linear_det(MuVar(input))

    # Overlay deterministic predictions on Monte Carlo samples.
    plt.hist(output_samples.detach().cpu().squeeze(), density=True)
    x = torch.linspace(output_samples.min(), output_samples.max(), 1000)
    plt.plot(
        x,
        (
            torch.exp(-0.5 * (x - output_det[0]) ** 2 / output_det[1])
            / torch.sqrt(2.0 * np.pi * output_det[1])
        )
        .detach()
        .cpu()
        .squeeze(),
        label="Analytical",
    )
    plt.plot(
        x,
        (
            torch.exp(-0.5 * (x - output_mc[0]) ** 2 / output_mc[1])
            / torch.sqrt(2.0 * np.pi * output_mc[1])
        )
        .detach()
        .cpu()
        .squeeze(),
        label="Monte Carlo",
    )
    plt.legend()
    plt.show()



"""Functionality for simulating observations of random variables."""

from collections.abc import Callable
from typing import Union

import numpy as np
import torch


def add_read_noise(
    signal: Union[float, np.array], sigma: Union[float, np.array]
) -> Union[float, np.array]:
    """Noisy realization of `signal` (read noise).

    Args:
        signal: Clean signal to which we add zero-mean Normal read noise.
        sigma: Standard deviation of zero-mean Normally distributed read noise. Can be
            homoscedastic (scalar) or heteroscedastic (array matching len(signal)).
    """
    return signal + sigma * np.float32(np.random.randn(*signal.shape))


def sensor_noise(
    signal: Union[float, np.array], sigma: Union[float, np.array]
) -> Union[float, np.array]:
    """Noisy realization of `signal` (read noise + shot noise).

    Args:
        signal: Clean signal to which we add read noise and shot noise.
        sigma: Standard deviation of zero mean Normally distributed read noise. Can be
            homoscedastic (scalar) or heteroscedastic (array matching len(signal)).
    """
    return add_read_noise(signal=np.float32(np.random.poisson(signal)), sigma=sigma)


class NoiseTransform(torch.nn.Module):
    """Wrapper to facilitate using noise functions with torch transform functionality."""

    def __init__(
        self,
        noise_fxn: Callable,
        noise_fxn_kwargs: dict = {},
        noise_fxn_kwargs_generator: dict = {},
    ) -> None:
        """Initializer for stochastic simulator dataset.

        Args:
            noise_fxn: Callable that noises an input signal.
            noise_fxn_kwargs: Keyword arguments passed to noise_fxn as
                noise_fxn(x, **noise_fxn_kwargs)
            noise_fxn_kwargs_generator: Keyword argument generator whose values
                can be called on the forward(x) pass to generate x-dependent
                kwargs that override noise_fxn_kwargs.  For example,
                if noise_fxn accepts an input argument `sigma` that can be a
                len(x) array, we can generate heteroscedastic (x-dependent)
                noise defined by `sigma` as
                noise_fxn_kwargs_generator={"sigma": lambda x: 0.1 * x**2}
        """
        super().__init__()
        self.noise_fxn = noise_fxn
        self.noise_fxn_kwargs = noise_fxn_kwargs
        self.noise_fxn_kwargs_generator = noise_fxn_kwargs_generator

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass to generate noisy `x`."""
        # Generate x-dependent arguments and merge with noise_fxn_kwargs.
        generated_kwargs = {
            key: value_gen(x)
            for key, value_gen in self.noise_fxn_kwargs_generator.items()
        }
        noise_fxn_kwargs = self.noise_fxn_kwargs | generated_kwargs
        return torch.tensor(self.noise_fxn(x, **noise_fxn_kwargs))


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import polynomials

    x = torch.linspace(-1.0, 1.0, 100)
    signal = polynomials.polynomial(x, order=1, coefficients=[0.0, 1.0])

    # Homoscedastic noise:
    noise_tform = NoiseTransform(
        noise_fxn=add_read_noise, noise_fxn_kwargs={"sigma": 0.1}
    )
    fig, ax = plt.subplots()
    ax.plot(x, signal, color="r", linewidth=2, label="clean signal")
    for _ in range(5):
        ax.plot(
            x,
            noise_tform(x),
            color="k",
            alpha=0.2,
        )
    ax.plot([], color="k", alpha=0.2, label="noisy realizations")
    ax.legend()
    plt.show()

    # Heteroscedastic noise generated from input x:
    noise_tform = NoiseTransform(
        noise_fxn=add_read_noise,
        noise_fxn_kwargs_generator={
            "sigma": lambda x: 0.1 + 0.1 * (1.0 + np.sin(2.0 * np.pi * x))
        },
    )
    fig, ax = plt.subplots()
    ax.plot(x, signal, color="r", linewidth=2, label="clean signal")
    for _ in range(5):
        ax.plot(
            x,
            noise_tform(x),
            color="k",
            alpha=0.2,
        )
    ax.plot([], color="k", alpha=0.2, label="noisy realizations")
    ax.legend()
    plt.show()



"""Functionality for simulating image data."""

from collections.abc import Iterable
from typing import Union

import numpy as np
import scipy.stats
import torch


def gaussian_blobs(
    mu: np.array,
    sigma: np.array,
    amplitude: Union[float, np.array],
    im_size: Iterable,
) -> np.array:
    """Noise-free image of a (possibly unnormalized) 2D isotropic Gaussian mixture model.

    Generate a noise-free image of the sum of 2D isotropic Gaussians in the domain
    x in [-(im_size[1]-1)/2, (im_size[1]+1)/2] and y in [-(im_size[0]-1)/2, (im_size[0]+1)/2].

    Args:
        mu: Locations of individual Gaussians. [n_gaussians, y, x]
        sigma: Standard deviation(s) of the len(mu) Gaussians. [1 or n_gaussians, y, x]
        amplitude: Amplitude(s) of the len(mu) Gaussians.  [1 or n_gaussians]
        im_size: Size of the output image in pixels.  [y, x]
    """
    # Reshape inputs.
    if len(sigma) != len(mu):
        sigma = sigma * np.ones_like(mu)
    if isinstance(amplitude, float):
        amplitude = amplitude * np.ones(mu.shape[0])

    # Compute output image.
    y = np.arange(im_size[0]) - im_size[0] / 2 + 0.5
    x = np.arange(im_size[1]) - im_size[1] / 2 + 0.5
    out = np.zeros((1, *im_size), dtype=np.float32)
    for n in range(len(mu)):
        out += (
            2.0
            * np.pi
            * np.prod(sigma[n])
            * amplitude[n]
            * (
                scipy.stats.norm.pdf(y, loc=mu[n][0], scale=sigma[n][0])[:, None]
                @ scipy.stats.norm.pdf(x, loc=mu[n][1], scale=sigma[n][1])[None, :]
            )
        )

    return out


class GridSamples:
    def __init__(self, n_per_pixel: int = 10, im_size: Iterable = (8, 8)) -> None:
        """Initialize stateful location sampler"""
        self.n_per_pixel = n_per_pixel
        grid = torch.meshgrid(
            torch.arange(-im_size[0] // 2, im_size[0] // 2) + 0.5,
            torch.arange(-im_size[0] // 2, im_size[0] // 2) + 0.5,
            indexing="ij",
        )
        self.xy = torch.stack((grid[0].flatten(), grid[1].flatten())).T.repeat(
            (n_per_pixel, 1)
        )
        self._counter = 0

    def __len__(self) -> int:
        """Return sampler length."""
        return len(self.xy)

    def __call__(self) -> torch.tensor:
        """Return xy location."""
        xy = self.xy[self._counter % len(self)][None, :]
        self._counter += 1
        return xy


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    im = gaussian_blobs(
        mu=np.array([[0.0, 0.0], [-1.0, 3.0]]),
        sigma=np.array([[0.7, 1.0], [0.5, 0.5]]),
        amplitude=np.array([1.0, 0.7]),
        im_size=(8, 8),
    )
    fig, ax = plt.subplots()
    plt.imshow(im, cmap="gray", extent=([-3.5, 3.5, 3.5, -3.5]))
    plt.show()

    im = gaussian_blobs(
        mu=np.array([[0.0, 0.0]]),
        sigma=np.array([[1.0, 1.0]]),
        amplitude=np.array([1.0]),
        im_size=(8, 8),
    )
    print(im.max())



"""Losses and helpers useful for Bayesian neural network training/evaluation."""

from abc import ABC, abstractmethod
from typing import Optional, Union

import torch
import torch.distributions as dist
from torch.nn.modules.loss import _Loss

from . import priors


def kl_divergence_sampled(
    dist0: dist.Distribution, dist1: dist.Distribution, n_samples: int = 1
):
    """KL divergence KL(dist0||dist1) approximated by Monte Carlo sampling."""
    kl_divergence = []
    for _ in range(n_samples):
        sample = dist0.sample()
        kl_divergence.append(dist0.log_prob(sample) - dist1.log_prob(sample))

    return torch.stack(kl_divergence).mean(dim=0)


class KLDivergence(_Loss):
    """KL divergence loss for Bayesian neural networks."""

    def __init__(self, prior: Optional[Union[dict, priors.Distribution]] = None):
        """Initialize KL divergence loss.
        Args:
            prior: Prior distribution over parameters.  This can be a single
                distribution for all parameters or a dictionary of dictionaries whose
                primary keys correspond to named modules and whose secondary keys
                correspond to parameters of that module.  The list indices
                correspond to model.named_modules(). By default, None
                will use priors set within each Bayesian layer on
                initialization.
        """
        super().__init__()
        self.prior = prior

    def forward(
        self,
        model: torch.nn.Module,
    ) -> torch.Tensor:
        """Compute KL divergence for Bayesian sub-modules of `model`.

        Args:
            model: torch.nn.Module that may have some Bayesian layers
                as sub-modules, for which we'll compute the KL divergence w.r.t
                their prior.
        """
        kl = []
        for module in model.named_modules():
            module_parameters = [p for p in module[1].parameters() if p.requires_grad]
            if (
                hasattr(module[1], "compute_kl_divergence")
                and (len(module_parameters) > 0)
                and module[1].learn_var
            ):
                if isinstance(self.prior, dict):
                    # Pass the input prior dictionary for this module.
                    kl.append(
                        module[1].compute_kl_divergence(priors=self.prior[module[0]])
                    )
                else:
                    kl.append(module[1].compute_kl_divergence(priors=self.prior))

        return torch.stack(kl).sum()


class BNNLoss(ABC, _Loss):
    """Abstract class for ELBO-like losses used to train Bayesian Neural Networks."""

    @property
    @abstractmethod
    def beta(self) -> torch.FloatTensor:
        """Scale factor for KL divergence loss term."""
        pass

    @property
    @abstractmethod
    def neg_log_likelihood(self) -> _Loss:
        """_Loss to compute the negative log-likelihood term in the ELBO."""
        pass

    @property
    @abstractmethod
    def kl_divergence(self) -> _Loss:
        """_Loss to compute the KL divergence of a model."""
        pass


class ELBO(BNNLoss):
    """Evidence lower bound with scaled KL."""

    def __init__(
        self,
        neg_log_likelihood: Optional[_Loss] = None,
        kl_divergence: _Loss = KLDivergence(),
        beta: float = 1.0,
        reduction: str = "sum",
    ) -> None:
        """Initialize ELBO loss.

        Args:
            neg_log_likelihood: Initialized neg_log_likelihood loss (e.g.,
                torch.nn.GaussianNLLLoss()).  This will be called in the forward
                pass of this loss as neg_log_likelihood(**kwargs) where **kwargs are
                the keyword arguments passed as ELBO()(**kwargs).
            kl_divergence: Initialized kl_divergence loss whose forward pass
                takes a torch.nn.Module `model` as input and returns a tensor
                corresponding to the KL divergence between parameters of `model`
                and their prior distribution.
            beta: Scaling parameter for KL loss term in the ELBO.
        """
        super().__init__(reduction=reduction)

        # Set default log likelihood calculator.
        if neg_log_likelihood is None:
            neg_log_likelihood = torch.nn.GaussianNLLLoss(
                full=True, reduction=reduction
            )
        self._neg_log_likelihood = neg_log_likelihood

        self._kl_divergence = kl_divergence
        self._beta = beta

    @property
    def beta(self) -> bool:
        """Return property `beta`.

        This property is written as an @property method for compatibility with the
        abstract parent class.
        """
        return self._beta

    @property
    def neg_log_likelihood(self) -> bool:
        """Return property `neg_log_likelihood`.

        This property is written as an @property method for compatibility with the
        abstract parent class.
        """
        return self._neg_log_likelihood

    @property
    def kl_divergence(self) -> bool:
        """Return property `kl_divergence`.

        This property is written as an @property method for compatibility with the
        abstract parent class.
        """
        return self._kl_divergence

    def forward(
        self, model: Optional[torch.nn.Module] = None, **kwargs
    ) -> torch.Tensor:
        """Compute the ELBO loss.

        Args:
            model: torch.nn.Module that may have some layers.BayesianLayers
                as sub-modules, for which we'll compute the KL divergence w.r.t
                their prior.  Passing None is treated as no model, i.e., KL = 0.0
            kwargs: Keyword arguments to pass to self.log_likelihood(**kwargs)
        """
        if (model is None) or (self.beta == 0.0) or (self.kl_divergence is None):
            return self.neg_log_likelihood(**kwargs)
        else:
            return self.neg_log_likelihood(**kwargs) + self.beta * self.kl_divergence(
                model
            )



"""Helpers for statistical analysis of BNNs."""

import torch


def compute_coverage(
    observations: torch.Tensor,
    mu: torch.Tensor = torch.tensor(0.0),
    sigma: torch.Tensor = torch.tensor(1.0),
    alphas: torch.Tensor = torch.tensor([1.0]),
) -> torch.Tensor:
    """Compute the coverage of `alpha` confidence intervals of a Gaussian."""
    coverage = []
    for alpha in alphas:
        in_interval = (observations - mu).abs() <= (sigma * alpha)
        coverage.append(in_interval.sum() / in_interval.numel())

    return torch.stack(coverage)



"""Functionality for simulating polynomial data."""

from collections.abc import Iterable
from typing import Union

import torch


def polynomial(
    x: Union[float, torch.Tensor], coefficients: Iterable = [0.0, 1.0]
) -> Union[float, torch.Tensor]:
    """Basic polynomial.

    Args:
        x: Points at which to evaluate the polynomial.
        coefficients: Polynomial coefficients in ascending order.
    """
    return torch.stack([c * (x**n) for n, c in enumerate(coefficients)]).sum(dim=0)


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    x = torch.linspace(-1.0, 1.0, 1000)
    fig, ax = plt.subplots()
    ax.plot(x, polynomial(x))
    plt.show()



"""Bayesian neural network base module(s) and utilities."""

import copy
from functools import partial
from typing import Any, Callable, Iterator, Union

import lightning as L
import torch

from .losses import BNNLoss
from .types import MuVar
from .wrappers import convert_to_bnn_


class BNN(torch.nn.Module):
    """Bayesian neural network base class."""

    def __init__(
        self,
        nn: torch.nn.Module,
        convert_in_place: bool = False,
        *args,
        **kwargs,
    ):
        """Initialize Bayesian neural network.


        WARNINGS:
            (1): Some functionality of this class relies on parameter names
                containing the suffixes "_mean" and "_rho".  If the input `nn`
                has parameters containing these strings, this class may not
                behave as expected!
            (2): The forward pass of `nn` is assumed to accept a single tensor
                representing the input.  The conversion to a BNN will hijack
                the forward pass through `nn` by changing the type of this
                input tensor to bnn.types.MuVar.

        Args:
            nn: Neural network to be converted to a Bayesian neural network.
            convert_in_place: Flag indicating input `nn` should be converted to
                a BNN in place.
            args, kwargs: Passed as
                bnn.utils.convert_to_bnn_(model=nn, *args, **kwargs)
        """
        super().__init__()

        # Convert the neural network to a Bayesian neural network.
        if not convert_in_place:
            nn = copy.deepcopy(nn)
        convert_to_bnn_(model=nn, *args, **kwargs)
        self.bnn = nn

    def named_parameters_tagged(self, tag: str) -> Iterator:
        """Return named parameters whose name contains `tag`."""
        for name, param in self.named_parameters():
            if tag in name:
                yield name, param

    def forward(self, input: Union[MuVar, torch.Tensor], *args, **kwargs) -> Any:
        """Forward pass through BNN."""
        return self.bnn(input, *args, **kwargs)


class BNNLightning(L.LightningModule):
    """PyTorch Lightning wrapper for BNN class."""

    def __init__(
        self,
        bnn: BNN,
        loss: BNNLoss,
        optimizer: Callable = partial(torch.optim.AdamW, lr=1.0e-3),
    ) -> None:
        """Initialize Lightning wrapper.

        Args:
            bnn: Bayesian neural network to wrap in Lightning.
            loss: Loss function to call in training/validation.
            optimizer: Partially initialized optimizer that will be given parameters
                to optimize in self.configure_optimizers().
        """
        super().__init__()

        self.bnn = bnn
        self.loss = loss
        self.optimizer_fxn = optimizer

    def forward(self, *args, **kwargs) -> Any:
        """Forward pass through BNN."""
        return self.bnn(*args, **kwargs)

    def configure_optimizers(self):
        return self.optimizer_fxn(self.parameters())

    def training_step(self, batch, batch_idx):
        """Training step for a single batch."""
        # Compute forward pass through model.
        out = self.bnn(MuVar(batch[0]))

        # Compute loss.
        loss = self.loss(model=self.bnn, input=out[0], target=batch[1], var=out[1])

        # Log results.
        self.log("train_loss", loss, prog_bar=True, sync_dist=True)

        return loss

    def validation_step(self, batch, batch_idx):
        """Validation step for a single batch."""
        # Compute forward pass through model.
        out = self.bnn(MuVar(batch[0]))

        # Compute loss.
        loss = self.loss(model=self.bnn, input=out[0], target=batch[1], var=out[1])

        # Log results.
        self.log("val_loss", loss, prog_bar=True, sync_dist=True)

        return loss


if __name__ == "__main__":
    from models import mlp

    # Convert a model to a Bayesian counterpart.
    in_features = 3
    out_features = 1
    model = mlp.MLP(
        in_features=in_features,
        out_features=out_features,
        n_hidden_layers=3,
        activation=torch.nn.LeakyReLU,
    )
    convert_to_bnn_(model=model)
    out = model(torch.ones(1, in_features))

    # Create a BNN wrapper for our model.
    model = mlp.MLP(
        in_features=in_features,
        out_features=out_features,
        n_hidden_layers=3,
        activation=torch.nn.LeakyReLU,
    )
    bnn = BNN(nn=model)
    bnn(torch.randn((1, 3)))



"""Generic dataset for simulated data."""

from collections.abc import Callable
from typing import Optional

import torch

from ..simulation import generators, observation


class SimulatedData(torch.utils.data.Dataset):
    """Dataset for stochastic simulators with fixed input arguments."""

    def __init__(
        self,
        data_generator: Callable,
        dataset_length: int = 1,
        cache: bool = False,
        transform: Optional[torch.nn.Module] = None,
    ) -> None:
        """Initializer for stochastic simulator dataset.

        Args:
            data_generator:  Callable that can be called to generate data as data = data_generator()
            dataset_length: Desired dataset length (set manually since `simulator` is stochastic).
            cache: Flag indicating data should be cached in memory (as opposed to sampling new
                data every __getitem__()).
            transform: Transform applied to simulated data before returning.
        """
        super().__init__()
        self.data_generator = data_generator
        self._len = dataset_length
        self.cache = cache
        self.transform = transform
        self.data = [None for _ in range(dataset_length)]

    def __len__(self) -> int:
        return self._len

    def __getitem__(self, idx: int) -> dict:
        if self.cache and self.data[idx] is not None:
            return self.data[idx]
        else:
            data_dict = self.data_generator()
            if self.transform is not None:
                data_dict["output"] = self.transform(data_dict["output"])
            if self.cache:
                self.data[idx] = data_dict

        return data_dict


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import numpy as np

    from simulation import generators, images, observation

    # Example dataset: simulated noisy images.
    data_generator = generators.Generator(
        simulator=images.gaussian_blobs,
        simulator_kwargs={
            "sigma": np.array([[1.0, 1.0]]),
            "amplitude": 100.0,
            "im_size": (8, 8),
        },
        simulator_kwargs_generator={
            "mu": lambda: 8 * (np.random.random(size=(1, 2)) - 0.5),
        },
    )
    simulator_kwargs = {"mu": np.array([[]])}
    noise_tform = observation.NoiseTransform(
        noise_fxn=observation.add_read_noise,
        noise_fxn_kwargs={"sigma": 0.1},
    )
    n_data = 3
    ds = SimulatedData(
        data_generator=data_generator, dataset_length=n_data, transform=noise_tform
    )
    for n in range(n_data):
        data = ds[n]
        fig, ax = plt.subplots()
        plt.imshow(
            data["output"],
        )
        plt.show()



"""Torch module for basic polynomials."""

from typing import Union

import torch

from fastbnns.bnn import types


class PolyModule(torch.nn.Module):
    """Polynomial model."""

    def __init__(self, poly_order: int = 1) -> None:
        """Initialize module."""
        super().__init__()
        self.poly_order = poly_order
        self.coeffs = torch.nn.Parameter(
            torch.randn(poly_order + 1, dtype=torch.float32),
            requires_grad=True,
        )

    def forward(self, input: Union[torch.Tensor, types.MuVar]) -> torch.Tensor:
        """Forward pass through module."""
        return torch.stack([c * (input**n) for n, c in enumerate(self.coeffs)]).sum(
            dim=0
        )



"""Miscellaneous utility functions useful throughout repository."""

import torch


def set_requires_grad_(module: torch.nn.Module, requires_grad: bool, tag: str) -> None:
    """Set requires_grad property of all parameters whose name contains `tag`."""
    for param_name, param_value in module.named_parameters():
        if tag in param_name:
            param_value.requires_grad = requires_grad



"""Collections of data generators to, e.g., aid dataset creation."""

from collections.abc import Callable
from typing import Any

import torch


def sample_simulator(
    simulator: Callable, simulator_kwargs_generator: dict, simulator_kwargs: dict
) -> dict:
    """Call simulator with randomly sampled inputs.

    Args:
        simulator: Callable that accepts keyword arguments and returns simulated
            data.
        simulator_kwargs_generator: Dictionary whose keys define keyword
            arguments of `simulator` and whose values are Callable and return
            valid values of associated keyword arguments.
        simulator_kwargs: Fixed keyword arguments to be merged with arguments
            generated by `simulator_kwargs_generator` before passing to `simulator`.
    """
    generated_kwargs = {
        key: value_gen() for key, value_gen in simulator_kwargs_generator.items()
    }
    simulator_kwargs = simulator_kwargs | generated_kwargs

    return {"input": simulator_kwargs, "output": simulator(**simulator_kwargs)}


class Generator(torch.nn.Module):
    """Generator to allow calling simulator with stochastic inputs."""

    def __init__(
        self,
        simulator: Callable,
        simulator_kwargs: dict,
        simulator_kwargs_generator: dict,
    ) -> None:
        """Initializer for stochastic simulator dataset.

        Args:
            simulator: Callable that accepts keyword arguments and returns simulated
                data.
            simulator_kwargs: Fixed keyword arguments to be merged with arguments
                generated by `simulator_kwargs_generator` before passing to `simulator`.
            simulator_kwargs_generator: Dictionary whose keys define keyword
                arguments of `simulator` and whose values are Callable and return
                valid values of associated keyword arguments.
        """
        super().__init__()
        self.simulator = simulator
        self.simulator_kwargs = simulator_kwargs
        self.simulator_kwargs_generator = simulator_kwargs_generator

    def forward(self) -> Any:
        return sample_simulator(
            simulator=self.simulator,
            simulator_kwargs=self.simulator_kwargs,
            simulator_kwargs_generator=self.simulator_kwargs_generator,
        )


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import numpy as np

    import polynomials

    # Example usage of sample_simulator(): simulate random samples from a polynomial.
    simulator = polynomials.polynomial
    simulator_kwargs_generator = {"x": lambda: np.random.random() - 0.5}
    simulator_kwargs = {"order": 2, "coefficients": (0.0, 0.0, 1.0)}
    sim_gen = Generator(
        simulator=simulator,
        simulator_kwargs=simulator_kwargs,
        simulator_kwargs_generator=simulator_kwargs_generator,
    )
    n_samples = 100
    x = np.empty(n_samples)
    y = np.empty(n_samples)
    for n in range(n_samples):
        data_dict = sim_gen()
        x[n] = data_dict["input"]["x"]
        y[n] = data_dict["output"]

    fig, ax = plt.subplots()
    ax.plot(x, y, ".")
    plt.show()



"""Simple CNN PyTorch models."""

from typing import Union

import torch


class CNN(torch.nn.Module):
    """Basic multi-layer CNN."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        hidden_features: int = 8,
        n_hidden_layers: int = 3,
        kernel_size: int = 3,
        stride: int = 1,
        padding: Union[int, str] = "same",
        activation: type = torch.nn.LeakyReLU,
    ):
        super().__init__()

        modules = [
            torch.nn.Conv2d(
                in_channels=in_channels,
                out_channels=hidden_features,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
            ),
        ]
        if activation is not None:
            modules.append(activation())
        for _ in range(n_hidden_layers):
            modules.append(
                torch.nn.Conv2d(
                    in_channels=hidden_features,
                    out_channels=hidden_features,
                    kernel_size=kernel_size,
                    stride=stride,
                    padding=padding,
                )
            )
            if activation is not None:
                modules.append(activation())
        modules.append(
            torch.nn.Conv2d(
                in_channels=hidden_features,
                out_channels=out_channels,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
            )
        )
        self.module_list = torch.nn.ModuleList(modules)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        for layer in self.module_list:
            x = layer(x)
        return x


if __name__ == "__main__":
    # Basic CNN usage example.
    in_channels = 1
    out_channels = 1
    network = CNN(in_channels=in_channels, out_channels=out_channels)
    batch_size = 8
    out = network(torch.ones((batch_size, in_channels, 128, 128)))



"""Test functionality of custom activations."""

import torch

from fastbnns.models.activations import InverseTransformSampling, scaled_sigmoid


def test_scaled_sigmoid() -> None:
    """Test functionality of scaled_sigmoid."""
    # Test some values to make sure the right function is applied.
    x = torch.tensor([0.0, 1.0, 2.0])
    alpha = torch.tensor([1.0, 2.0, 3.0])
    y_expected = torch.tensor([0.5000, 0.8808, 0.9975])
    y = scaled_sigmoid(x=x, alpha=alpha)
    assert (torch.abs(y_expected - y) < 1.0e-3).all(), (
        "Output of `scaled_sigmoid` not matching expected results!"
    )


def test_inverse_tform_activation() -> None:
    "Test basic functionality of InverseTransformSampling custom activation."
    # Test basic usage of activation.
    activation = InverseTransformSampling(
        alpha_init=torch.tensor([1.0]),
        distribution=torch.distributions.Normal(loc=0.0, scale=1.0),
    )
    x = torch.tensor([0.0000, 0.4001, 1.2276])
    y_expected = torch.tensor([0.0000, 0.2500, 0.7501])
    y = activation(x)
    assert (torch.abs(y_expected - y) < 1.0e-3).all(), (
        "Output of `InverseTransformSampling` forward pass not matching expected results!"
    )



"""Inference modules for Bayesian neural network layers.

Custom propagators for specific layers (e.g., "Linear" for Bayesian analog of
torch.nn.Linear) should share a name with the layer such that
getattr(inference, layer.__class__.__name__) will return the custom propagator
for that layer if available.
"""

from collections.abc import Callable, Iterable
from typing import List, Optional, Union

import torch
import torch.distributions as dist


class MomentPropagator(torch.nn.Module):
    """Base class for layer propagators."""

    def __init__(self):
        """Initializer for MomentPropagator module."""
        super().__init__()

    def forward(
        self,
        module: Callable,
        input: Iterable,
    ) -> Union[Iterable, tuple]:
        """Forward method for MomentPropagator modules.

        Args:
            module: Instance of the layer through which we will propagate moments.
            input: Input passed to layer, which will typically be a torch.Tensor or
                types.MuVar.
        """
        raise NotImplementedError


class BasicPropagator(MomentPropagator):
    """Propagate mean and variance through `module`.

    This propagator can be used with modules that have "simple" forward passes
    for which propagation rules are already defined by methods in types.MuVar
    (e.g., forward pass is just input*param1 + param2).
    """

    def __init__(self):
        """Initializer for BasicPropagator inference module."""
        super().__init__()

    def forward(
        self,
        module: Callable,
        input: Iterable,
    ) -> Union[Iterable, tuple]:
        """Propagate moments by relying intrinsically on methods in types.MuVar."""
        return module.module(input)


class UnscentedTransform(MomentPropagator):
    """Unscented transform propagation of mean and variance through `module`.

    This propagator uses the unscented transform to propagate mean and variance
    through a deterministic layer.
    """

    def __init__(
        self,
        sigma_scale: Optional[torch.Tensor] = None,
        sigma_weights: Optional[torch.Tensor] = None,
        n_module_samples: int = 1,
    ):
        """Initializer for UnscentedTransform inference module.

        Args:
            sigma_scale: Scale factor for sigma points mu -+ sigma_scale*var.sqrt().
            sigma_weights: Weighting factors corresponding to sigma points.
            n_module_samples: Number of samples to make of the module itself.
                This allows us to use the unscented transform through modules
                which themselves are parametrized by random variables.
        """
        super().__init__()

        # Set defaults as needed (defaults chosen as in https://doi.org/10.1117/12.280797)
        kappa = 2
        if sigma_scale is None:
            sigma_scale = torch.sqrt(torch.tensor(kappa + 1))
        self.register_buffer("_scale", sigma_scale)

        if sigma_weights is None:
            sigma_weights = torch.tensor(
                [kappa / (kappa + 1), 0.5 / (kappa + 1), 0.5 / (kappa + 1)]
            )
        self.register_buffer("_weights", sigma_weights)
        self._weights = sigma_weights

        self.n_module_samples = n_module_samples

    def forward(
        self,
        module: Callable,
        input: Iterable,
        return_samples: bool = False,
    ) -> Union[Iterable, tuple]:
        """Propagate moments using the unscented transform."""
        # Select sigma points and reshape along batch dimension for batched eval.
        scaled_stdev = self._scale * input[1].sqrt()
        sigma_points = torch.stack(
            (input[0], input[0] - scaled_stdev, input[0] + scaled_stdev)
        )
        sp_shape = sigma_points.shape
        sigma_points = sigma_points.reshape(sp_shape[0] * sp_shape[1], *sp_shape[2:])
        weights = self._weights

        # Propagate mean and variance.
        if self.n_module_samples > 1:
            # Perform n_module_samples unscented transforms and combine results.
            mu_samples = []
            var_samples = []
            for n in range(self.n_module_samples):
                # Prepare a sampled instance of the module.  For stochastic modules,
                # module.module returns a new sample of weights each time, so
                # we need to prepare the instance before running .forward() on each
                # sigma point.
                if hasattr(module, "module"):
                    module_sample = module.module
                else:
                    module_sample = module

                # Forward pass through module and use unscented transform.
                samples = module_sample(sigma_points)
                samples = samples.reshape(sp_shape[0], sp_shape[1], *samples.shape[1:])
                if n == 0:
                    weights = weights.reshape(
                        (weights.shape[0],) + (1,) * (samples.ndim - 1)
                    )
                mu_samples.append((weights * samples).sum(dim=0))
                var_samples.append(
                    (weights * ((samples - mu_samples[-1]) ** 2)).sum(dim=0)
                )

            # Combine estimates from each unscented transform using law of total
            # expectation and law of total variance.
            mu = torch.stack(mu_samples).mean(dim=0)
            var = torch.stack(var_samples).mean(dim=0) + torch.stack(mu_samples).var(
                dim=0
            )
        else:
            # Prepare a sampled instance of the module.  For stochastic modules,
            # module.module returns a new sample of weights each time, so
            # we need to prepare the instance before running .forward() on each
            # sigma point.
            if hasattr(module, "module"):
                module_sample = module.module
            else:
                module_sample = module

            # Compute output mean and variance.
            samples = module_sample(sigma_points)
            samples = samples.reshape(sp_shape[0], sp_shape[1], *samples.shape[1:])
            weights = weights.reshape((weights.shape[0],) + (1,) * (samples.ndim - 1))
            mu = (weights * samples).sum(dim=0)
            var = (weights * ((samples - mu) ** 2)).sum(dim=0)

        if return_samples:
            return type(input)([mu, var]), samples
        else:
            return type(input)([mu, var])


class MonteCarlo(MomentPropagator):
    """Monte Carlo propagation of mean and variance through `module`.

    This propagator is designed to perform Monte Carlo sampling of BOTH
    the input and the layer being propagated through.  This would generally
    only be used for Bayesian/stochastic layers through which we cannot
    propagate moments through analytically.
    """

    def __init__(
        self, n_samples: int = 10, input_sampler: dist.Distribution = dist.Normal
    ):
        """Initializer for MonteCarlo inference module.

        Args:
            n_samples: Number of Monte Carlo samples used to estimate mean
                and variance after passing through this layer.
            input_sampler: Distribution template used to sample the input to
                a layer in the forward pass as
                input_sample = input_sampler(loc=input_mu, scale=input_var.sqrt()).sample()

        """
        super().__init__()
        self.n_samples = n_samples
        self.input_sampler = input_sampler

    def forward(
        self,
        module: Callable,
        input: Iterable,
        return_samples: bool = False,
    ) -> Union[Iterable, tuple]:
        """Propagate moments by averaging over n_samples forward passes of module."""
        # If the input variance is greater than zero, we'll need to sample the
        # input as well.
        if (input[1] > 0.0).all():
            input_dist = self.input_sampler(loc=input[0], scale=input[1].sqrt())
            samples = torch.stack(
                [module(input_dist.sample()) for _ in range(self.n_samples)]
            )
        else:
            samples = torch.stack([module(input[0]) for _ in range(self.n_samples)])

        if return_samples:
            return type(input)([samples.mean(dim=0), samples.var(dim=0)]), samples
        else:
            return type(input)([samples.mean(dim=0), samples.var(dim=0)])


class Linear(MomentPropagator):
    """Deterministic moment propagation of mean and variance through a Linear layer."""

    functional = torch.nn.functional.linear

    def __init__(self):
        """Initializer for Linear inference module"""
        super().__init__()

    def forward(
        self,
        module: torch.nn.Module,
        input: Iterable,
    ) -> Iterable:
        """Analytical moment propagation through layer."""
        ## Compute analytical result under mean-field approximation following
        ## https://doi.org/10.48550/arXiv.2402.14532
        # Reorganize parameters.
        module_params = module._module_params
        bias_mean = module_params.get("bias_mean", None)
        bias_rho = module_params.get("bias_rho", None)
        weight_mean = module_params["weight_mean"]
        weight_rho = module_params["weight_rho"]

        # Propagate mean.
        mu = self.functional(
            input=input[0],
            weight=weight_mean,
            bias=bias_mean,
        )

        # Propagate variance: first term accounts for parameter variance, second
        # term propagates input variance.
        if weight_rho is None:
            # No "first term" since there is no parameter variance.
            var = self.functional(
                input=input[1],
                weight=weight_mean**2,
            )
        else:
            weight_var = module.scale_tform(weight_rho) ** 2
            var = self.functional(
                input=input[0] ** 2,
                weight=weight_var,
                bias=(None if bias_rho is None else module.scale_tform(bias_rho) ** 2),
            ) + self.functional(
                input=input[1],
                weight=weight_mean**2 + weight_var,
            )

        return type(input)([mu, var])


class ConvNd(MomentPropagator):
    """Deterministic moment propagation of mean and variance through a ConvNd layer."""

    def __init__(self):
        """Initializer for ConvNd inference module"""
        super().__init__()

    def forward(
        self,
        module: torch.nn.Module,
        input: Iterable,
    ) -> Iterable:
        """Analytical moment propagation through layer."""
        # Modify input and prepare functional arguments.
        if module._module.padding_mode != "zeros":
            # This is added to mimic torch.nn.modules.conv _conv_forward methods.
            input = torch.nn.functional.pad(
                input,
                module._module._reversed_padding_repeated_twice,
                mode=module._module.padding_mode,
            )
            functional_kwargs = {
                "stride": module._module.stride,
                "padding": torch.nn.modules.utils._triple(0),
                "dilation": module._module.dilation,
                "groups": module._module.groups,
            }
        else:
            functional_kwargs = {
                "stride": module._module.stride,
                "padding": module._module.padding,
                "dilation": module._module.dilation,
                "groups": module._module.groups,
            }

        ## Compute analytical result under mean-field approximation following
        ## https://doi.org/10.48550/arXiv.2402.14532
        # Reorganize parameters.
        module_params = module._module_params
        bias_mean = module_params.get("bias_mean", None)
        bias_rho = module_params.get("bias_rho", None)
        weight_mean = module_params["weight_mean"]
        weight_rho = module_params["weight_rho"]

        # Propagate mean.
        mu = self.functional(
            input=input[0],
            weight=weight_mean,
            bias=bias_mean,
            **functional_kwargs,
        )

        # Propagate variance: first term accounts for parameter variance, second
        # term propagates input variance.
        if weight_rho is None:
            # No "first term" since there is no parameter variance.
            var = self.functional(
                input=input[1],
                weight=weight_mean**2,
                bias=None,
                **functional_kwargs,
            )
        else:
            weight_var = module.scale_tform(weight_rho) ** 2
            var = self.functional(
                input=input[0] ** 2,
                weight=weight_var,
                bias=(None if bias_rho is None else module.scale_tform(bias_rho) ** 2),
                **functional_kwargs,
            ) + self.functional(
                input=input[1],
                weight=weight_mean**2 + weight_var,
                bias=None,
                **functional_kwargs,
            )

        return type(input)([mu, var])


class Conv1d(ConvNd):
    """Deterministic moment propagation of mean and variance through a Conv1d layer.

    The internal logic is identical to ConvNd so we just create this class for compatibility
    with module-name-based searches.
    """

    functional = torch.nn.functional.conv1d

    def __init__(self):
        """Initializer for Conv1d inference module"""
        super().__init__()


class Conv2d(ConvNd):
    """Deterministic moment propagation of mean and variance through a Conv2d layer.

    The internal logic is identical to ConvNd so we just create this class for compatibility
    with module-name-based searches.
    """

    functional = torch.nn.functional.conv2d

    def __init__(self):
        """Initializer for Conv2d inference module"""
        super().__init__()


class Conv3d(ConvNd):
    """Deterministic moment propagation of mean and variance through a Conv3d layer.

    The internal logic is identical to ConvNd so we just create this class for compatibility
    with module-name-based searches.
    """

    functional = torch.nn.functional.conv3d

    def __init__(self):
        """Initializer for Conv3d inference module"""
        super().__init__()


class ConvTransposeNd(MomentPropagator):
    """Deterministic moment propagation of mean and variance through a ConvTransposeNd layer."""

    def __init__(self):
        """Initializer for ConvTransposeNd inference module"""
        super().__init__()

    def forward(
        self,
        module: torch.nn.Module,
        input: Iterable,
        output_size: Optional[List[int]] = None,
    ) -> Iterable:
        """Analytical moment propagation through layer."""
        # Prepare functional arguments.
        output_padding = module._module._output_padding(
            input,
            output_size,
            module._module.stride,
            module._module.padding,
            module._module.kernel_size,
            self.num_spatial_dims,
            module._module.dilation,
        )
        functional_kwargs = {
            "stride": module._module.stride,
            "padding": module._module.padding,
            "dilation": module._module.dilation,
            "groups": module._module.groups,
            "output_padding": output_padding,
        }

        ## Compute analytical result under mean-field approximation following
        ## https://doi.org/10.48550/arXiv.2402.14532
        # Reorganize parameters.
        module_params = module._module_params
        bias_mean = module_params.get("bias_mean", None)
        bias_rho = module_params.get("bias_rho", None)
        weight_mean = module_params["weight_mean"]
        weight_rho = module_params["weight_rho"]

        # Propagate mean.
        mu = self.functional(
            input=input[0],
            weight=weight_mean,
            bias=bias_mean,
            **functional_kwargs,
        )

        # Propagate variance: first term accounts for parameter variance, second
        # term propagates input variance.
        if weight_rho is None:
            # No "first term" since there is no parameter variance.
            var = self.functional(
                input=input[1],
                weight=weight_mean**2,
                bias=None,
                **functional_kwargs,
            )
        else:
            weight_var = module.scale_tform(weight_rho) ** 2
            var = self.functional(
                input=input[0] ** 2,
                weight=weight_var,
                bias=(None if bias_rho is None else module.scale_tform(bias_rho) ** 2),
                **functional_kwargs,
            ) + self.functional(
                input=input[1],
                weight=weight_mean**2 + weight_var,
                bias=None,
                **functional_kwargs,
            )

        return type(input)([mu, var])


class ConvTranspose1d(ConvTransposeNd):
    """Deterministic moment propagation of mean and variance through a ConvTranspose1d layer."""

    functional = torch.nn.functional.conv_transpose1d

    def __init__(self):
        """Initializer for ConvTranspose1d inference module"""
        super().__init__()
        self.num_spatial_dims = 1


class ConvTranspose2d(ConvTransposeNd):
    """Deterministic moment propagation of mean and variance through a ConvTranspose2d layer."""

    functional = torch.nn.functional.conv_transpose2d

    def __init__(self):
        """Initializer for ConvTranspose2d inference module"""
        super().__init__()
        self.num_spatial_dims = 2


class ConvTranspose3d(ConvTransposeNd):
    """Deterministic moment propagation of mean and variance through a ConvTranspose3d layer."""

    functional = torch.nn.functional.conv_transpose3d

    def __init__(self):
        """Initializer for ConvTranspose3d inference module"""
        super().__init__()
        self.num_spatial_dims = 3


if __name__ == "__main__":
    """Example usages of inference modules."""
    import matplotlib.pyplot as plt

    # Define a nonlinearity to propagate through.
    layer = torch.nn.LeakyReLU()

    # Define some propagators.
    mc = MonteCarlo(n_samples=100)
    ut = UnscentedTransform()

    # Propagate example data through layer.
    input = (torch.tensor([1.23])[None, :], torch.tensor([3.21])[None, :])
    out_mc, samples_mc = mc(module=layer, input=input, return_samples=True)
    out_ut, samples_ut = ut(module=layer, input=input, return_samples=True)

    # Plot results.
    x = torch.linspace(
        (input[0] - 3.0 * torch.sqrt(input[1])).squeeze(),
        (input[0] + 3.0 * torch.sqrt(input[1])).squeeze(),
        1000,
    )
    fig, ax = plt.subplots()
    ax.hist(samples_mc.squeeze(), density=True, label="Monte Carlo samples")
    ax.plot(
        x,
        torch.distributions.Normal(
            loc=out_mc[0].squeeze(), scale=out_mc[1].squeeze().sqrt()
        )
        .log_prob(x)
        .exp(),
        label="Monte Carlo estimated PDF",
    )
    ax.plot(
        x,
        torch.distributions.Normal(
            loc=out_ut[0].squeeze(), scale=out_ut[1].squeeze().sqrt()
        )
        .log_prob(x)
        .exp(),
        label="Unscented transform estimated PDF",
    )
    plt.legend()
    plt.show()



"""Custom activation functions."""

import torch
from torch.distributions import Distribution


def scaled_sigmoid(
    x: torch.Tensor, alpha: torch.Tensor = torch.tensor(1.0)
) -> torch.Tensor:
    """Compute the scaled sigmoid function \frac{1.0}{1.0+\exp(-\alpha*x)}"""
    return 1.0 / (1.0 + torch.exp(-alpha * x))


class InverseTransformSampling(torch.nn.Module):
    """Activation to mimic inverse transform sampling from some distribution."""

    def __init__(
        self,
        distribution: Distribution = torch.distributions.Normal(loc=0.0, scale=1.0),
        learn_alpha: bool = False,
        alpha_init: torch.tensor = torch.tensor([1.0]),
        eps: float = 1.0e-6,
        *args,
        **kwargs
    ) -> None:
        """Initialize InverseTransformSampling class.

        Args:
            distribution: Torch distribution with a defined .icdf() method.
            learn_alpha: Flag indicating we should learn the alpha scale in
                the domain transform f(x) = 1.0 / (1.0 + exp(-alpha*x)),
                otherwise alpha=1.0 will always be used.
            eps: Clamp inputs to distribution.icdf to [eps, 1.0-eps]
        """
        super().__init__(*args, **kwargs)

        # Define the domain transform to convert inputs in
        # (-\inf, \inf) to [0, 1]
        self._alpha = torch.nn.Parameter(
            torch.log(torch.exp(alpha_init) - 1.0), requires_grad=learn_alpha
        )  # self.alpha=softplus(self._alpha)
        self.domain_tform = scaled_sigmoid

        # Define the Normal distribution of interest.
        self.distribution = distribution
        self.eps = eps

    @property
    def alpha(self) -> torch.Tensor:
        """Scale self._alpha to ensure positivity and return."""
        return torch.nn.functional.softplus(self._alpha)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass through activation."""
        # Transform inputs from (-\inf, \inf) to [eps, 1.0-eps]
        x_prime = self.domain_tform(x, alpha=self.alpha)
        x_prime.clamp_(min=self.eps, max=1.0 - self.eps)

        # Treat transformed inputs as samples from U[0, 1] and pass through
        # inverse CDF of self.disribution.
        return self.distribution.icdf(x_prime)



"""Custom types and associated functionality."""

from __future__ import annotations
import functools
from typing import Any, Callable, Optional, Union

import math
import torch


# List torch functions that we can apply independently to mean and variance.
SIMPLE_TORCH_FUNCS = [
    torch.cat,
    torch.chunk,
    torch.dsplit,
    torch.column_stack,
    torch.dstack,
    torch.gather,
    torch.hsplit,
    torch.hstack,
    torch.index_select,
    torch.masked_select,
    torch.movedim,
    torch.permute,
    torch.reshape,
    torch.select,
    torch.split,
    torch.stack,
    torch.take_along_dim,
    torch.tensor_split,
    torch.tile,
    torch.transpose,
    torch.unbind,
    torch.unravel_index,
    torch.squeeze,
    torch.unsqueeze,
    torch.vsplit,
    torch.vstack,
    torch.where,
    torch.sum,  # assumes independence: E[a+b]=E[a]+E[b], V[a+b]=V[a]+V[b]
    torch.detach,
    torch.clone,
    torch.nn.functional.pad,
    torch.nn.functional.interpolate,
    torch.nn.functional.upsample,
    torch.nn.functional.upsample_nearest,
    torch.nn.functional.upsample_bilinear,
    torch.nn.functional.grid_sample,
    torch.nn.functional.affine_grid,
]

# Define additional tensor-specific methods that can only be called as x.method(), not torch.method(x).
TENSOR_METHODS = ["cpu", "cuda", "to", "requires_grad_", "view"]


class MuVar:
    """Custom list-like object holding mean and variance of some distribution.

    WARNING: Some functionality, like __pow__(), assumes the normal distribution!"""

    def __init__(
        self,
        mu: Union[
            torch.Tensor,
            list[torch.Tensor, torch.Tensor],
            tuple[torch.Tensor, torch.Tensor],
            MuVar,
        ],
        var: Optional[torch.Tensor] = None,
    ) -> None:
        """Initialize MuVar instance.

        Args:
            mu: Mean of input distribution, or optionally, a list containing
                both mean and variance (to allow calling this method on a list
                without unpacking arguments).
            var: Variance of input distribution.
        """
        if isinstance(mu, (list, tuple)):
            # Mean and variance passed as a list or tuple. Ensure we store as
            # a list so that it is mutable.
            self.mu_var = list(mu)
        elif isinstance(mu, MuVar):
            # Repackage for compatibility.
            self.mu_var = mu.mu_var
        elif var is None:
            # Only mean was passed, default variance to None
            # (which will be treated as zero when possible/appropriate).
            self.mu_var = (mu, None)
        else:
            # Mu and var passed individually.
            self.mu_var = (mu, var)

    @classmethod
    def __torch_function__(
        self,
        func: Callable,
        types: list,
        args: Any = (),
        kwargs: dict = {},
    ) -> Any:
        """General overloading function for torch functions."""

        # Ensure this is the __torch_function__ we need to call.
        # See https://pytorch.org/docs/stable/notes/extending.html
        if not any(issubclass(t, MuVar) for t in types):
            return NotImplemented

        # For "simple" functions like torch.cat, we'll route mu and var separately
        # through the function.  Otherwise, we don't want to use these
        # __torch_function__ implementations.
        if func in SIMPLE_TORCH_FUNCS:

            def split_muvar(args: Union[list, tuple]) -> Union[list, tuple]:
                """Recursively split MuVar instances into separate lists of mu and lists of var."""
                if isinstance(args, Union[list, tuple]):
                    # Loop through arguments and split.
                    args_mu = []
                    args_var = []
                    for arg in args:
                        if isinstance(arg, MuVar):
                            # Split into mu and var, explicitly materializing None var's to zeros
                            # (to enable generic function applications expecting a tensor valued var).
                            args_mu.append(arg[0])
                            args_var.append(
                                torch.zeros_like(arg[0]) if arg[1] is None else arg[1]
                            )
                        elif isinstance(arg, (list, tuple)):
                            # Split MuVar items if needed.
                            splits = split_muvar(arg)
                            args_mu.append(splits[0])
                            args_var.append(splits[1])
                        else:
                            # Append to both argument lists.
                            args_mu.append(arg)
                            args_var.append(arg)
                    return args_mu, args_var
                else:
                    # This is for, e.g., scalar arguments.
                    return args, args

            # Separate MuVar types into mu and var for split calls to `func`.
            args_mu, args_var = split_muvar(args)
            kwargs_mu = {}
            kwargs_var = {}
            for k, v in kwargs.items():
                kwargs_mu[k], kwargs_var[k] = split_muvar(v)
            return MuVar(func(*args_mu, **kwargs_mu), func(*args_var, **kwargs_var))
        elif hasattr(self, func.__name__):
            # A custom implementation of this torch function was defined for this type.
            return getattr(self, func.__name__)(*args, **kwargs)
        else:
            # Return NotImplemented to allow other overrides to be used.
            return NotImplemented

    def __repr__(self):
        """Custom display functionality."""
        return f"MuVar({self.mu_var})"

    def __getattr__(self, name: str) -> Any:
        """Custom getattr fallback handler."""
        # If a torch function exists with name `name` (e.g., x.sum()), return that.
        # Otherwise we'll return the requested attribute for self.mu_var[0].
        torch_fxn = getattr(torch, name, None)
        if name in TENSOR_METHODS:
            # These methods can only be called as tensor.method() but otherwise can be applied
            # to mu and var independently.
            return functools.partial(getattr(self, "apply_method"), name)
        elif (torch_fxn is not None) and callable(torch_fxn):
            return functools.partial(torch_fxn, self)
        else:
            mu = self.mu_var[0]
            if hasattr(mu, name):
                return getattr(mu, name)

        raise AttributeError(f"'{type(self).__name__}' has no attribute '{name}'")

    def __getitem__(self, idx: int) -> torch.Tensor:
        """Access requested index of self.mu_var"""
        return self.mu_var[idx]

    def __add__(self, input: Union[float, torch.Tensor, list, MuVar]) -> MuVar:
        """Custom add functionality for MuVar types."""
        if isinstance(input, (float, torch.Tensor)):
            # Adding a float or tensor is like adding a delta R.V., so
            # variance does not change.
            return MuVar(self.mu_var[0] + input, self.mu_var[1])
        elif isinstance(input, (list, MuVar)):
            # Means and variances both add (assuming independent distributions).
            if (self.mu_var[1] is None) and (input[1] is None):
                return MuVar(self.mu_var[0] + input[0], None)
            elif self.mu_var[1] is None:
                return MuVar(self.mu_var[0] + input[0], input[1])
            elif input[1] is None:
                return MuVar(self.mu_var[0] + input[0], self.mu_var[1])
            else:
                return MuVar(self.mu_var[0] + input[0], self.mu_var[1] + input[1])
        else:
            raise NotImplementedError

    def __radd__(self, input: Union[float, torch.Tensor, list]) -> MuVar:
        """Custom add functionality for MuVar types."""
        return self.__add__(input)

    def __sub__(self, input: Union[float, torch.Tensor, list, MuVar]) -> MuVar:
        """Custom subtract functionality for MuVar types."""
        if isinstance(input, (float, torch.Tensor)):
            # Adding a float or tensor is like adding a delta R.V., so
            # variance does not change.
            return MuVar(self.mu_var[0] - input, self.mu_var[1])
        elif isinstance(input, (list, MuVar)):
            # Means can be subtracted but variances still add.
            if (self.mu_var[1] is None) and (input[1] is None):
                return MuVar(self.mu_var[0] - input[0], None)
            elif self.mu_var[1] is None:
                return MuVar(self.mu_var[0] - input[0], input[1])
            elif input[1] is None:
                return MuVar(self.mu_var[0] - input[0], self.mu_var[1])
            else:
                return MuVar(self.mu_var[0] - input[0], self.mu_var[1] + input[1])
        else:
            raise NotImplementedError

    def __rsub__(self, input: Union[float, torch.Tensor, list]) -> MuVar:
        """Custom subtract functionality for MuVar types."""
        return self.__sub__(input)

    def __mul__(self, input: Union[float, torch.Tensor, list, MuVar]) -> MuVar:
        """Custom multiply functionality for MuVar types."""
        if isinstance(input, (float, torch.Tensor)):
            # Multiplication by scalar: E[aX] = aE[x], V[aX]=a**2 V[X]
            if self.mu_var[1] is None:
                return MuVar(input * self.mu_var[0], None)
            else:
                return MuVar(input * self.mu_var[0], (input**2) * self.mu_var[1])
        elif isinstance(input, (list, MuVar)):
            # Multiplication of two random independent variables:
            #   E[XY] = E[X]E[Y]
            #   V[XY] = (E[X]**2)*V[Y] + V[X]*(E[Y]**2) + V[X]*V[Y]
            mu = input[0] * self.mu_var[0]
            if (self.mu_var[1] is None) and (input[1] is None):
                var = None
            elif self.mu_var[1] is None:
                var = (self.mu_var[0] ** 2) * input[1]
            elif input[1] is None:
                var = self.mu_var[1] * (input[0] ** 2)
            else:
                var = (self.mu_var[0] ** 2) * input[1] + self.mu_var[1] * (
                    (input[0] ** 2) + input[1]
                )
            return MuVar(mu, var)
        else:
            raise NotImplementedError

    def __rmul__(self, input: Union[float, torch.Tensor, list]) -> MuVar:
        """Custom multiply functionality for MuVar types."""
        return self.__mul__(input)

    def __matmul__(self, input: Union[torch.Tensor, list, MuVar]) -> MuVar:
        """Custom matrix multiply functionality for MuVar types."""
        # NOTE: MuVar is NOT holding multivariate distributions.  Each scalar entry
        # represents (mu, var) of an independent distribution, so matrix multiplication
        # is not multiplication of multivariate random variables!
        if isinstance(input, torch.Tensor):
            # Multiplication by scalar: E[a@X] = a @ E[x], V[a@X]=a**2 @ V[X]
            if self.mu_var[1] is None:
                return MuVar(input @ self.mu_var[0], None)
            else:
                return MuVar(input @ self.mu_var[0], input @ self.mu_var[1] @ input.T)
        elif isinstance(input, (list, MuVar)):
            # Multiplication of two random independent variables:
            #   E[X@Y] = E[X] @ E[Y]
            #   V[X@Y] = E[X]**2 @ V[Y] + V[X] @  E[Y]**2 + V[X] @ V[Y]
            mu = self.mu_var[0] @ input[0]
            if (self.mu_var[1] is None) and (input[1] is None):
                var = None
            elif self.mu_var[1] is None:
                var = (self.mu_var[0] ** 2) @ input[1]
            elif input[1] is None:
                var = self.mu_var[1] @ (input[0] ** 2)
            else:
                var = (self.mu_var[0] ** 2) @ input[1] + self.mu_var[1] @ (
                    (input[0] ** 2) + input[1]
                )
            return MuVar(mu, var)
        else:
            raise NotImplementedError

    def __rmatmul__(self, input: Union[torch.Tensor, list]) -> MuVar:
        """Custom matrix multiply functionality for MuVar types."""
        return self.__matmul__(input)

    def __pow__(self, input: int) -> MuVar:
        """Custom exponentiation functionality for MuVar types.

        WARNING: This implementation assumes independent Normally distributed random variables!
        """
        if isinstance(input, int):
            # Exponentiation of a Normal random variable: see
            # https://en.wikipedia.org/wiki/Normal_distribution#Moments
            def normal_moment(
                mu: torch.Tensor,
                v: torch.Tensor,
                n: int,
            ) -> torch.Tensor:
                """Compute E[X^n] for X ~ Normal(mu, v)."""
                moment = torch.zeros_like(mu)
                for m in range(n // 2 + 1):
                    coeff = math.comb(n, 2 * m) * math.prod(range(1, 2 * m, 2))
                    moment += coeff * (v**m) * (mu ** (n - 2 * m))
                return moment

            mu = normal_moment(mu=self.mu_var[0], v=self.mu_var[1], n=input)
            if self.mu_var[1] is None:
                var = None
            else:
                var = (
                    normal_moment(mu=self.mu_var[0], v=self.mu_var[1], n=2 * input)
                    - normal_moment(mu=self.mu_var[0], v=self.mu_var[1], n=input) ** 2
                )
            return MuVar(mu, var)
        else:
            raise NotImplementedError

    def apply(self, func: Callable, *args, **kwargs) -> MuVar:
        """Generic apply() for functions that act separately on mu and var."""
        if self.mu_var[1] is None:
            # If self.mu_var[1] is None (zero variance), we need to explicitly materialize the
            # zeros tensor to accommodate arbitrary func().
            return MuVar(
                func(self.mu_var[0], *args, **kwargs),
                func(torch.zeros_like(self.mu_var[0]), *args, **kwargs),
            )
        else:
            return MuVar(
                func(self.mu_var[0], *args, **kwargs),
                func(self.mu_var[1], *args, **kwargs),
            )

    def apply_method(self, name: str, *args, **kwargs) -> MuVar:
        """Generic apply_method() for methods that act separately on mu and var."""
        if self.mu_var[1] is None:
            # If self.mu_var[1] is None (zero variance), we need to explicitly materialize the
            # zeros tensor to accommodate arbitrary func().
            return MuVar(
                getattr(self.mu_var[0], name)(*args, **kwargs),
                getattr(torch.zeros_like(self.mu_var[0]), name)(*args, **kwargs),
            )
        else:
            return MuVar(
                getattr(self.mu_var[0], name)(*args, **kwargs),
                getattr(self.mu_var[1], name)(*args, **kwargs),
            )

    def numel(self) -> int:
        """Custom numel() to avoid complicated logic in __getattr__ above."""
        return self.mu_var[0].numel()

    def mean(self, *args, **kwargs) -> MuVar:
        """Custom replacement of torch.mean() for MuVar type."""
        # Assuming independence, we can directly apply mean to self.mu_var[0].
        x_mean = self.mu_var[0].mean(*args, **kwargs)

        # To compute variance, we call the torch version of mean() with
        # keepdim=True so we can account for the scaling factor:
        # assuming independence, V[(x_0+x_1) / 2] = (V[x_0]+V[x_1]) / 4
        if self.mu_var[1] is None:
            x_var = None
        else:
            if (args == ()) and (kwargs == {}):
                x_var = self.mu_var[1].mean()
                x_var /= self.mu_var[1].numel() ** 2
            else:
                x_var = self.mu_var[1].mean(*args, **(kwargs | {"keepdim": True}))
                x_var /= (
                    torch.tensor(self.mu_var[1].shape) / torch.tensor(x_var.shape)
                ).prod().squeeze() ** 2
        if kwargs.pop("keepdim", None) is None:
            return MuVar(x_mean, x_var)
        else:
            return MuVar(x_mean.squeeze(), None if x_var is None else x_var.squeeze())


if __name__ == "__main__":
    # Scalar operations.
    a = MuVar(torch.tensor(1.0), torch.tensor(2.0))
    b = MuVar(torch.tensor(1.1), torch.tensor(0.5))

    print(a + b)
    print(a + 1.0)
    print(1.0 + a)
    print(a**2)

    # Torch/tensor operations and attributes.
    a = MuVar(torch.randn((2, 2)), torch.ones((2, 2)))
    b = MuVar(torch.randn((2, 2)), 1.1 * torch.ones((2, 2)))

    print(a.shape)
    print(a.size(1))
    print(a.to("cpu"))
    print(a.numel())
    print(a.sum())
    print(a.mean())
    print(a.mean(dim=1, keepdim=True))
    print(a @ b)
    print(torch.cat([a, b], dim=-1))
    print(torch.nn.functional.pad(a, [0, 1, 2, 0]))
